Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun tuple ssralg matrix mxalgebra bigop.
Require Import reduction_definition inverse_nilpotent hexagonal_lemma.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section BPL.

Open Local Scope ring_scope.

Section decomposition.

Variable K : fieldType.
Variables (c0 c1 d0 d1 : nat).
Variables (D1 : 'M[K]_(c1,c0)) (D1' : 'M[K]_(d1,d0))
          (F0 : 'M[K]_(c0,d0)) (F1 : 'M[K]_(c1,d1)) 
          (G0 : 'M[K]_(d0,c0)) (G1 : 'M[K]_(d1,c1)) 
          (H0 : 'M[K]_(c0,c1)).

Hypothesis F0_morph : (D1 *m F0) = (F1 *m D1').
Hypothesis G0_morph : (D1' *m G0) = (G1 *m D1).
Hypothesis ax1_0 : (G0 *m F0) = 1%:M.
Hypothesis ax1_1 : (G1 *m F1) = 1%:M.
Hypothesis ax2_0 : (F0 *m G0) + (H0 *m D1) = 1%:M.
Hypothesis ax2_1 : (F1 *m G1) + (D1 *m H0) = 1%:M.
Hypothesis ax3_0 : (H0 *m F1) = 0.
Hypothesis ax4_0 : (G0 *m H0) = 0.



Section auxiliary.

Lemma cast_rank1 (m n : nat) (M : 'M[K]_(m,n)) : 
 ((\rank M) + (m - (\rank M))%N )%N = m. 
Proof.
by rewrite (subnKC (rank_leq_row M)).
Qed.

Lemma ltnSadd (m n : nat) : m + n < m = false.
Proof.
elim : n => [|n IH]; first by rewrite addn0 ltnn.
rewrite addnS.
case H : ((m + n).+1 < m); last by done.
have H2 : m + n < (m + n).+1 by done.
move : (ltn_trans H2 H) => H3.
by rewrite H3 in IH.  
Qed.


Definition ker_min  (m n : nat) (M : 'M_(m,n)) :=
  (castmx ((Logic.eq_refl (m-\rank M)%N), (cast_rank1 M))
          (row_mx (@const_mx _ (m-\rank M) (\rank M) 0) 
                  1%:M)) *m 
  (kermx M).


Lemma castmx_copid (m m' r: nat) (eq_mm : (m = m') * (m = m'))
  (R : ringType) :
  castmx eq_mm (copid_mx r) = (@copid_mx R m' r).
Proof. by case: eq_mm; case: m' / => b; rewrite castmx_id. Qed.


Lemma castmx_blockmx_copid (m m' r: nat) (eq_mm : ((r + m')%N = m) * ((r+m')%N = m))
  (R : ringType) : (castmx eq_mm (block_mx 0 0 0 1%:M)) = (@copid_mx R m r).
Proof.
case : eq_mm; case : m / => b; rewrite castmx_id.
apply/matrixP => i j; rewrite mxE; case : splitP => i0 H;
rewrite mxE; case : splitP => j0 Hj; 
rewrite !mxE {2}H ?(ltn_ord i0) ?Bool.andb_true_r ?ltnSadd 
        ?Bool.andb_false_r ?GRing.subr0 /=. 
      case  Hij : (@eq_op nat_eqType (@nat_of_ord (addn r m') i)
                    (@nat_of_ord (addn r m') j)).
         move : Hij; move/eqP => Hij;  move : (ord_inj Hij); 
         move/eqP => -> /=.
         have H11 : 1%:R == 1%:R by done.
         move : (H11 R) => H12.      
         rewrite -GRing.subr_eq0 in H12.
         by rewrite (eqP H12).
       case Hij2 : (i == j); last by rewrite GRing.subr0.
       move : (eqP Hij2) => Hij3.
       by rewrite Hij3 in Hij; move : Hij; move/eqP.
    case  Hij : (@eq_op nat_eqType (@nat_of_ord (addn r m') i)
                    (@nat_of_ord (addn r m') j)).
       move : Hij; move/eqP => Hij;  move : (ord_inj Hij); 
       move/eqP => -> /=.
       have H11 : 1%:R == 1%:R by done.
       move : (H11 R) => H12.      
       rewrite -GRing.subr_eq0 in H12.
       by rewrite (eqP H12).
     case Hij2 : (i == j); last by rewrite GRing.subr0.
     move : (eqP Hij2) => Hij3.
     by rewrite Hij3 in Hij; move : Hij; move/eqP.
  case Hij: (i == j); last by done.
  move : (eqP Hij) => Hij2.
  rewrite Hij2 in H.
  by move : (ltn_ord j0); rewrite -Hj H ltnSadd.
case Hij: (i0 == j0). 
  move : (eqP Hij) => Hij2.
  rewrite Hij2 in H.
  have Hij3 : (@eq_op nat_eqType (@nat_of_ord (addn r m') i)
                    (@nat_of_ord (addn r m') j)) by rewrite H Hj.
  move: (ord_inj (eqP Hij3)) => ->.
  by have -> : (j == j); by done. 
case Hij2 : (i == j); last by done.
move : (eqP Hij2) => Hij3.
rewrite Hij3 Hj in H.
have Hij4 : (r + j0 - r)%N = (r + i0 - r)%N by rewrite H. 
rewrite -{2 4}(addn0 r) !subn_add2l !subn0 in Hij4.
move : (ord_inj Hij4) => Hij5.
rewrite Hij5 in Hij.
by move : Hij; move/eqP.
Qed.


Lemma mulmx_cast : forall (R : ringType) (m1 n1 m2 n2 m3 m4 : nat) (eq_mn1 : (m1 = m2) * (n1 = n2))
     (eq_mn2 : (n1 = n2) * (m3 = m4)) (eq_mn3 : (m1 = m2) * (m3 = m4)) (A : 'M[R]_(m1, n1)) (B :'M[R]_(n1,m3)), 
     ((castmx eq_mn1 A) *m (castmx eq_mn2 B)) = (castmx eq_mn3 (A *m B)).
Proof.
move => R m1 n1 m2 n2 m3 m4 [eq_mn1 eq_mn1'] [eq_mn2 eq_mn2'] [eq_mn3 eq_mn3'] A B.
move: (eq_mn1) (eq_mn1') (eq_mn2) (eq_mn2') (eq_mn3) (eq_mn3') A B. 
rewrite eq_mn1 eq_mn1' eq_mn3' => eq_mn_0 eq_mn_1 eq_mn_2 eq_mn_3 eq_mn_4 eq_mn_5 A B.
by rewrite !castmx_id.
Qed.

Lemma ker_min_kermx (m n : nat) (M : 'M_(m,n)) : 
  (kermx M :=: (ker_min M))%MS. 
Proof.
apply/eqmxP/andP; split; apply/submxP; last first.
  by exists (castmx ((Logic.eq_refl (m-\rank M)%N), (cast_rank1 M))
                    (row_mx (@const_mx _ (m-\rank M) (\rank M) 0)  1%:M)).
exists (castmx ((cast_rank1 M),(Logic.eq_refl (m-\rank M)%N))
               (col_mx (@const_mx _ (\rank M) (m-\rank M)%N 0) 1%:M)).
rewrite /ker_min mulmxA mulmx_cast ?cast_rank1 //.
move => Hyp0.
by rewrite mul_col_row !mul0mx mulmx0 mulmx1 /kermx castmx_blockmx_copid 
           mulmxA copid_mx_id ?(rank_leq_row M). 
Qed.

End auxiliary.



Section isomorphism.

Definition castF0 := 
(castmx ((cast_rank1 F0), (Logic.eq_refl (c0-\rank F0)%N))
    (col_mx (@const_mx K (\rank F0) (c0-\rank F0)%N 0) 1%:M)).

Definition castF1 := 
(castmx ((cast_rank1 F1), (Logic.eq_refl (c1-\rank F1)%N))
    (col_mx (@const_mx K (\rank F1) (c1-\rank F1)%N 0)  1%:M)).

Definition F0_isom := 
   (col_mx (ker_min F0) G0).

Definition G0_isom :=
  (row_mx (H0 *m D1 *m (col_ebase F0) *m castF0) F0).

Definition F1_isom := 
   (col_mx (ker_min F1) G1).

Definition G1_isom := (row_mx ((D1 *m H0) *m (col_ebase F1) *m castF1) F1).



Section lemmas_isomorphism.

Lemma F0G0 : G0_isom *m F0_isom = 1%:M.
Proof.
rewrite /F0_isom /G0_isom !mul_row_col -mulmxA /castF0.
have -> : (castmx (cast_rank1 F0, erefl (c0 - \rank F0)%N)
  (col_mx (const_mx 0) 1%:M) *m ker_min F0) = kermx F0.
  rewrite /ker_min mulmxA mulmx_cast ?cast_rank1 // => Hyp0.
  by rewrite mul_col_row !mul0mx mulmx0 mulmx1 /kermx castmx_blockmx_copid 
             mulmxA copid_mx_id ?(rank_leq_row F0). 
have HH1 : H0 *m D1 *m F0 = 0; first by rewrite -mulmxA F0_morph mulmxA ax3_0 mul0mx. 
by rewrite (mulmxKV_ker HH1) GRing.addrC.
Qed.

Lemma G0F0 :  F0_isom *m G0_isom = 1%:M.
Proof.
rewrite /G0_isom /F0_isom !mul_col_row ax1_0 !mulmxA ax4_0 !mul0mx. 
move : (ker_min_kermx F0).
move/eqmxP; move/andP => [_ Hb].
move : Hb; move/submxP => [D' HD']; rewrite HD'.
rewrite -[_ *m F0]mulmxA mulmx_ker mulmx0.
have -> : (D' *m kermx F0 *m H0 *m D1 *m col_ebase F0 *m castF0) = 1%:M.
  rewrite -mulmxA -mulmxA -mulmxA [H0 *m _]mulmxA.
  have -> : (H0 *m D1) = 1%:M - (F0 *m G0).
    rewrite -ax2_0 [F0 *m _ + _]GRing.addrC -GRing.addrA.
    have H4 : F0 *m G0 == F0 *m G0 by done.
    rewrite -GRing.subr_eq0 in H4.
    by rewrite (eqP H4) GRing.addr0.
  rewrite !mulmx_addl !mulmx_addr mul1mx !mulNmx !mulmxN -mulmxA -mulmxA -mulmxA
  [kermx F0 *m (F0 *m _)]mulmxA mulmx_ker !mul0mx mulmx0 GRing.subr0 mulmxA -HD'
  /ker_min /kermx -!mulmxA mulmxA. 
  have -> : (invmx (col_ebase F0) *m (col_ebase F0 *m castF0)) = castF0; first 
    by rewrite mulmxA (mulVmx (col_ebase_unit F0)) mul1mx. 
  rewrite -(@castmx_blockmx_copid _ (c0 - \rank F0))
 ?(cast_rank1 F0) // => Hyp0; rewrite mulmx_cast ?(cast_rank1 F0) // => Hyp1; rewrite mulmx_cast ?(cast_rank1 F0) // => Hyp2.
  by rewrite mul_row_block mul_row_col mul0mx GRing.add0r mulmx0 GRing.add0r
          !mulmx1 mulmx0 GRing.add0r castmx_id.
by rewrite scalar_mx_block.
Qed.


Lemma F1G1 : G1_isom *m F1_isom = 1%:M.
Proof.
rewrite /F1_isom /G1_isom !mul_row_col -mulmxA /castF1 -mulmxA -mulmxA.
have -> : (castmx (cast_rank1 F1,  erefl (c1 - \rank F1)%N)
         (col_mx (const_mx 0) 1%:M) *m  ker_min  F1) =  kermx F1.
  rewrite /ker_min mulmxA mulmx_cast ?cast_rank1 // => Hyp0.
  by rewrite mul_col_row !mul0mx mulmx0 mulmx1 /kermx castmx_blockmx_copid 
             mulmxA copid_mx_id ?(rank_leq_row F1). 
have HH2 : D1 *m H0 *m F1 = 0 by rewrite -mulmxA ax3_0 mulmx0.
by rewrite mulmxA mulmxA (mulmxKV_ker HH2) GRing.addrC ax2_1.
Qed.


Lemma G1F1 :  F1_isom *m G1_isom = 1%:M.
Proof.
rewrite /G1_isom /F1_isom !mul_col_row ax1_1 mulmxA mulmxA. 
have -> : (G1 *m (D1 *m H0 *m col_ebase F1 *m castF1)) =
          (G1 *m D1 *m H0 *m col_ebase F1 *m castF1) by rewrite !mulmxA.
rewrite -G0_morph -[_ *m G0 *m _]mulmxA ax4_0 mulmx0 !mul0mx.
move : (ker_min_kermx F1).
move/eqmxP; move/andP => [_ Hb].
move : Hb; move/submxP => [D' HD']; rewrite HD'.
rewrite -[D' *m _ *m F1]mulmxA mulmx_ker !mulmx0.
have -> : (D' *m kermx F1 *m (D1 *m H0) *m col_ebase F1 *m castF1) = 1%:M.
  have -> : (D1 *m H0) = 1%:M - (F1 *m G1).
    by rewrite -ax2_1 -[F1 *m G1 + _ + _]GRing.addrA GRing.addrA
               [F1 *m G1 + _]GRing.addrC -GRing.addrA m_m_0 GRing.addr0. 
  rewrite !mulmx_addr !mulmx_addl mulmx1 !mulmxN !mulNmx.
  have -> : (D' *m kermx F1 *m (F1 *m G1)) *m col_ebase F1 *m castF1 = 0.
    by rewrite mulmxA -[D' *m _ *m F1]mulmxA  mulmx_ker !mulmx0 !mul0mx.
  rewrite GRing.subr0 -HD' /ker_min /kermx -!mulmxA mulmxA. 
  have -> : (invmx (col_ebase F1) *m (col_ebase F1  *m castF1)) = castF1. 
  by rewrite mulmxA (mulVmx (col_ebase_unit F1)) mul1mx. 
  rewrite -(@castmx_blockmx_copid _ (c1 - \rank F1)) ?(cast_rank1 F1) // => Hyp0; rewrite mulmx_cast ?(cast_rank1 F1) // => Hyp1; rewrite mulmx_cast ?(cast_rank1 F1) // => Hyp2.
  by rewrite mul_row_block mul_row_col mul0mx GRing.add0r mulmx0 GRing.add0r
          !mulmx1 mulmx0 GRing.add0r castmx_id.
by rewrite -scalar_mx_block.
Qed.

End lemmas_isomorphism.

Section new_reduction.

Definition D1_new :=  F1_isom *m D1 *m G0_isom.

Definition H0_new :=  F0_isom *m H0 *m G1_isom.

Definition F0_new := F0_isom *m F0.
Definition F1_new := F1_isom *m F1.

Definition G0_new := G0 *m G0_isom.
Definition G1_new := G1 *m G1_isom.

Section reduction_properties.

Lemma f0g0 : G0_new *m F0_new = 1%:M.
Proof.
by rewrite /G0_new /F0_new -mulmxA [G0_isom *m _]mulmxA F0G0 mul1mx ax1_0.
Qed.

Lemma f1g1 : G1_new *m F1_new = 1%:M.
Proof.
by rewrite -mulmxA [G1_isom *m _]mulmxA F1G1 mul1mx ax1_1.
Qed.

Lemma f1h0 : H0_new *m F1_new = 0.
Proof.
by rewrite /H0_new /F1_new -mulmxA [G1_isom *m _]mulmxA F1G1 mul1mx 
          -mulmxA ax3_0 mulmx0.
Qed.

Lemma h0g0 : G0_new *m H0_new = 0.
Proof.
by rewrite /G0_new /H0_new -mulmxA ![G0_isom *m _]mulmxA F0G0 mul1mx mulmxA
           ax4_0 mul0mx.
Qed.

Lemma g0f0_D1h0 : F0_new *m G0_new + H0_new *m D1_new = 1%:M.
Proof.
by rewrite /G0_new /H0_new /F0_new /D1_new -!mulmxA ![G1_isom *m _]mulmxA 
           F1G1 mul1mx -mulmx_addr !mulmxA -mulmx_addl ax2_0 mul1mx G0F0.
Qed.

Lemma g1f1_D2h1_h0D1 : 
  F1_new *m G1_new + D1_new *m H0_new = 1%:M.
Proof. 
rewrite /G1_new /F1_new /D1_new /H0_new -!mulmxA
        -mulmx_addr !mulmxA -mulmx_addl -!mulmxA.
have -> : D1 *m (G0_isom *m (F0_isom *m H0)) =
          D1 *m (G0_isom *m F0_isom) *m H0 by rewrite !mulmxA.
by rewrite F0G0 mulmx1 ax2_1 mul1mx G1F1.
Qed.

Lemma d1f1_f0D1 : F1_new *m D1' = D1_new *m F0_new.
Proof. 
by rewrite /D1_new /F0_new /F1_new -!mulmxA [G0_isom *m _]mulmxA F0G0 mul1mx F0_morph.
Qed.

Lemma D1g1_g0d1 : G1_new *m D1_new = D1' *m G0_new.
Proof. 
by rewrite /D1_new /G1_new /G0_new -!mulmxA [G1_isom *m _]mulmxA F1G1 
            mul1mx !mulmxA -G0_morph.
Qed.

End reduction_properties.

Section lemmas_for_hexagonal_lemma.

Definition H0_aux :=
  (ker_min F0 *m H0 *m (D1 *m H0 *m col_ebase F1 *m castF1)).

Lemma H0_newP :
  H0_new = block_mx H0_aux 0 0 0.
Proof.
by rewrite /H0_new !mul_col_mx mul_mx_row mul_mx_row /G1_isom ax4_0 !mul0mx
           -[_ *m F1]mulmxA ax3_0 mulmx0.
Qed.

Definition D1_aux :=
  (ker_min F1 *m D1 *m (H0 *m D1 *m col_ebase F0 *m castF0)).

Lemma D1_newP :
  D1_new = block_mx D1_aux 0 0 D1'.
Proof.
rewrite /D1_new !mul_col_mx mul_mx_row mul_mx_row /G0_isom.
have -> : (G1 *m D1 *m F0) = D1' by rewrite -mulmxA F0_morph mulmxA ax1_1 mul1mx.
have -> : (G1 *m D1 *m (H0 *m D1 *m col_ebase F0 *m castF0)) = 0
  by rewrite -G0_morph mulmxA mulmxA mulmxA -[_ *m G0 *m _]mulmxA ax4_0 
             mulmx0 !mul0mx.
have -> : (ker_min F1 *m D1 *m F0) = 0.
  rewrite -mulmxA F0_morph mulmxA.
  move : (ker_min_kermx F1).
  move/eqmxP; move/andP => [_ Hb].
  move : Hb; move/submxP => [D'' ->].
  by rewrite -[D'' *m _ *m _]mulmxA mulmx_ker !mulmx0 !mul0mx.
by rewrite -block_mxEv.
Qed.

Lemma aux1 :  F1_isom *m (F1 *m G1) *m G1_isom = 
              block_mx 0 0 0 1%:M.
Proof.
rewrite /G1_isom /F1_isom !mul_col_mx mulmxA. 
have -> : (ker_min F1 *m F1 *m G1 *m row_mx (D1 *m H0 *m col_ebase F1 *m
           castF1) F1) = 0.
  move : (ker_min_kermx F1).
  move/eqmxP; move/andP => [_ Hb].
  move : Hb; move/submxP => [D' HD']; rewrite HD'.
  by rewrite -[D' *m _ *m F1]mulmxA mulmx_ker !mulmx0 !mul0mx.
have -> : (G1 *m (F1 *m G1) *m row_mx (D1 *m H0 *m col_ebase F1 *m castF1) F1) = (row_mx 0 1%:M).
  by rewrite mulmxA ax1_1 mul1mx mul_mx_row ax1_1 !mulmxA -G0_morph
         -[D1' *m G0 *m _]mulmxA ax4_0 mulmx0 !mul0mx.
by rewrite -row_mx0 -block_mxEv.
Qed.

Lemma D1_H0_new_ulsubmx :
 (ulsubmx (D1_new *m H0_new)) = 1%:M.
Proof.
rewrite /D1_new /H0_new !mulmxA -[_ *m G0_isom *m _]mulmxA
        F0G0 mulmx1 -[_ *m D1 *m _]mulmxA.
have -> : (D1 *m H0) = 1%:M - (F1 *m G1).
    by rewrite -ax2_1 -[F1 *m G1 + _ + _]GRing.addrA GRing.addrA
               [F1 *m G1 + _]GRing.addrC -GRing.addrA m_m_0 GRing.addr0. 
  rewrite !mulmx_addr !mulmx_addl mulmx1 !mulmxN !mulNmx.
by rewrite G1F1 aux1 scalar_mx_block opp_block_mx add_block_mx GRing.subr0
        block_mxKul.
Qed.

Lemma D1_aux_H0_aux_1 :
 D1_aux *m H0_aux = 1%:M.
Proof.
move : D1_H0_new_ulsubmx.
by rewrite D1_newP H0_newP mulmx_block block_mxKul mul0mx GRing.addr0.
Qed.

Lemma H0_aux_D1_aux_1 :
 H0_aux *m D1_aux = 1%:M.
Proof.
have : H0_new *m D1_new = 1%:M - F0_new *m G0_new.
  rewrite -g0f0_D1h0 [F0_new *m _ + _]GRing.addrC -GRing.addrA.
  have H4 : F0_new *m G0_new == F0_new *m G0_new by done.
  rewrite -GRing.subr_eq0 in H4.
  by rewrite (eqP H4) GRing.addr0.
rewrite /F0_new /G0_new H0_newP D1_newP mulmx_block !mulmx0 !mul0mx !GRing.addr0.
have -> : F0_isom *m F0 *m (G0 *m G0_isom) = 
    block_mx 0 (ker_min F0 *m F0 *m (G0 *m F0))
     (G0 *m F0 *m (G0 *m (H0 *m D1 *m col_ebase F0 *m castF0)))
     (G0 *m F0 *m (G0 *m F0)).
  rewrite /F0_isom /G0_isom !mul_mx_row !mul_col_mx block_mxEh.
  have -> : (ker_min F0 *m F0 *m (G0 *m (H0 *m D1 *m col_ebase F0 *m castF0))) = 0.
    move : (ker_min_kermx F0).
    move/eqmxP; move/andP => [_ Hb].
    by move : Hb; move/submxP => [D' HD']; rewrite HD' -mulmxA -mulmxA
      [kermx F0 *m _]mulmxA mulmx_ker mul0mx mulmx0.
  by done.
rewrite scalar_mx_block opp_block_mx add_block_mx GRing.subr0 => H.
by move : (eq_block_mx H) => [-> _].
Qed.


End lemmas_for_hexagonal_lemma.

End new_reduction.

End isomorphism.



Section perturbation.

Variable P1 : 'M[K]_(c1, c0).


Variable (n : nat).

Hypothesis nilpotency_hp : (pot_matrix (P1 *m H0) n = 0).

Section new_perturbation.

Definition P1_new := F1_isom *m P1 *m G0_isom. 

Definition D1_new_pert := D1_new + P1_new.

Lemma pot_P1_new_H0_new k : pot_matrix (P1_new *m H0_new) k =
   F1_isom *m (pot_matrix (P1 *m H0) k) *m G1_isom.
Proof.
elim : k => [/=| k H]; first by rewrite mulmx1 G1F1.  
have -> : pot_matrix (P1_new *m H0_new) k.+1 = 
  (P1_new *m H0_new) *m pot_matrix (P1_new *m H0_new) k by done.
rewrite H /P1_new /H0_new -mulmxA -mulmxA -mulmxA -mulmxA -mulmxA
       [G0_isom *m _]mulmxA F0G0 mul1mx ![G1_isom *m _]mulmxA F1G1 mul1mx.
have -> : (P1 *m (H0 *m (pot_matrix (P1 *m H0) k *m G1_isom))) = 
          (P1 *m H0) *m (pot_matrix (P1 *m H0) k) *m G1_isom by rewrite !mulmxA.
have <- : pot_matrix (P1 *m H0) k.+1 = P1 *m H0 *m pot_matrix (P1 *m H0) k by done. 
by rewrite mulmxA. 
Qed.


Lemma nilpotency_hp_new : 
 pot_matrix (P1_new *m H0_new) n = 0.
Proof.
by rewrite pot_P1_new_H0_new nilpotency_hp mulmx0 mul0mx.
Qed.

Lemma pot_block_a_0_b_0 (p1 p2 k: nat) (a : 'M[K]_p1)
   (b : 'M[K]_(p2,p1)) : 
  ulsubmx (pot_matrix (block_mx a 0 b 0) k) = pot_matrix a k.
Proof.
elim : k =>  [/=| k H]; first by rewrite scalar_mx_block block_mxKul.
have -> : (pot_matrix (block_mx a 0 b 0) k.+1) =
 (block_mx a 0 b 0) *m (pot_matrix (block_mx a 0 b 0) k) by done.
by rewrite -(submxK (pot_matrix (block_mx a 0 b 0) k)) mulmx_block
        block_mxKul mul0mx H GRing.addr0.
Qed.

Lemma H0_aux_P11 : pot_matrix ((ulsubmx P1_new) *m H0_aux) n = 0.
Proof.
move : nilpotency_hp_new.
rewrite -{1}(submxK P1_new) H0_newP mulmx_block -block_mx0 !mulmx0
        !GRing.addr0 -(submxK (pot_matrix (block_mx (ulsubmx P1_new *m H0_aux) 
     0 (dlsubmx P1_new *m H0_aux) 0) n)).
move => H.
move : (eq_block_mx H) => [HH _ _ _].
by rewrite pot_block_a_0_b_0 in HH.
Qed.

Definition inverse_d1p1_11 :=  
 H0_aux *m (\sum_(0<=i<n) (pot_matrix (- ((ulsubmx P1_new) *m H0_aux)) i)).

Lemma inverse_d1p1_11_is_inverse :
  daa D1_new_pert *m inverse_d1p1_11 = 1%:M.  
Proof.
rewrite /D1_new_pert /daa -(submxK D1_new) -(submxK P1_new) add_block_mx
        block_mxKul -/D1_aux.
move : D1_newP.
rewrite -{1}(submxK D1_new) => HH.
move : (eq_block_mx HH) => [-> _ _ _].
by rewrite -(mulmx1 (ulsubmx P1_new))-{1}H0_aux_D1_aux_1 -{1}(mul1mx D1_aux)
        mulmxA -mulmx_addl /inverse_d1p1_11 -mulmxA [D1_aux *m _]mulmxA
        D1_aux_H0_aux_1 mul1mx (inverse_I_plus_M_big' H0_aux_P11).
Qed.

Lemma inverse_d1p1_11_is_inverse' :
  inverse_d1p1_11 *m daa D1_new_pert = 1%:M.  
Proof.
rewrite /D1_new_pert /daa -(submxK D1_new) -(submxK P1_new) add_block_mx
        block_mxKul -/D1_aux.
move : D1_newP.
rewrite -{1}(submxK D1_new) => HH.
move : (eq_block_mx HH) => [-> _ _ _].
by rewrite -(mulmx1 (ulsubmx P1_new))-{1}H0_aux_D1_aux_1 -{1}(mul1mx D1_aux)
        mulmxA -mulmx_addl /inverse_d1p1_11 -mulmxA 
        [_ *m (_ *m D1_aux)]mulmxA (inverse_I_plus_M_big H0_aux_P11)
        mul1mx H0_aux_D1_aux_1.
Qed.


Lemma inverse_d1p1_11_inverse : 
  inverse_d1p1_11 *m daa D1_new_pert = 1%:M 
  /\ daa D1_new_pert *m inverse_d1p1_11 = 1%:M. 
Proof.
by split; rewrite ?inverse_d1p1_11_is_inverse' ?inverse_d1p1_11_is_inverse.
Qed.

Definition quasi_bpl :=
  (rho inverse_d1p1_11_inverse).

Section definitions_BPL.

Definition D1_BPL := (D1 + P1).
Definition D1'_BPL := (reduction_definition.D1' quasi_bpl).

Definition F0_BPL := 
 G0_isom *m (reduction_definition.F0 quasi_bpl).
Definition F1_BPL :=
 G1_isom *m (reduction_definition.F1 quasi_bpl).

Definition G0_BPL := 
 (reduction_definition.G0 quasi_bpl) *m F0_isom.
Definition G1_BPL :=
 (reduction_definition.G1 quasi_bpl) *m F1_isom.

Definition H0_BPL := 
 G0_isom *m (reduction_definition.H0 quasi_bpl) *m F1_isom.

Section reduction_properties.

Lemma f0g0_BPL : G0_BPL *m F0_BPL = 1%:M.
Proof.
by rewrite /G0_BPL /F0_BPL -mulmxA [F0_isom *m _]mulmxA G0F0 mul1mx
        (reduction_definition.ax1_0 quasi_bpl). 
Qed.

Lemma f1g1_BPL : G1_BPL *m F1_BPL = 1%:M.
Proof.
by rewrite /G1_BPL /F1_BPL -mulmxA [F1_isom *m _]mulmxA G1F1 mul1mx
        (reduction_definition.ax1_1 quasi_bpl).
Qed.

Lemma f1h0_BPL : H0_BPL *m F1_BPL = 0.
Proof.
by rewrite /H0_BPL /F1_BPL -mulmxA [F1_isom *m _]mulmxA G1F1 mul1mx -mulmxA
       (reduction_definition.ax3_0 quasi_bpl) mulmx0.
Qed.

Lemma h0g0_BPL : G0_BPL *m H0_BPL = 0.
Proof.
by rewrite /H0_BPL /G0_BPL -mulmxA ![F0_isom *m _]mulmxA G0F0 mul1mx mulmxA
       (reduction_definition.ax4_0 quasi_bpl) mul0mx.
Qed.

Lemma g0f0_D1h0_BPL : F0_BPL *m G0_BPL + H0_BPL *m D1_BPL = 1%:M.
Proof.
rewrite /F0_BPL /G0_BPL /H0_BPL /D1_BPL.
have -> : (D1 + P1) = G1_isom *m (reduction_definition.D1 quasi_bpl) *m F0_isom
  by rewrite /= /D1_new_pert /D1_new /P1_new -mulmx_addl -mulmx_addr 
            -!mulmxA F0G0 mulmx1 mulmxA F1G1 mul1mx.
by rewrite !mulmxA -mulmx_addl -!mulmxA -mulmx_addr [F1_isom *m _]mulmxA 
           G1F1 mul1mx (reduction_definition.ax2_0 quasi_bpl) mulmx1 F0G0.
Qed.

Lemma g1f1_D2h1_h0D1_BPL : 
  F1_BPL *m G1_BPL + D1_BPL *m H0_BPL = 1%:M.
Proof. 
rewrite /F1_BPL /G1_BPL /H0_BPL /D1_BPL.
have -> : (D1 + P1) = G1_isom *m (reduction_definition.D1 quasi_bpl) *m F0_isom
  by rewrite /= /D1_new_pert /D1_new /P1_new -mulmx_addl -mulmx_addr 
            -!mulmxA F0G0 mulmx1 mulmxA F1G1 mul1mx.
by rewrite !mulmxA -!mulmx_addl -!mulmxA -!mulmx_addr [F0_isom *m _]mulmxA 
           G0F0 !mul1mx (reduction_definition.ax2_1 quasi_bpl) mulmx1 F1G1.
Qed.

Lemma d1f1_f0D1_BPL : F1_BPL *m D1'_BPL = D1_BPL *m F0_BPL.
Proof. 
rewrite /F1_BPL /D1'_BPL /D1_BPL /F0_BPL -mulmxA
        -(reduction_definition.F0_morph quasi_bpl).
have -> : (D1 + P1) = G1_isom *m (reduction_definition.D1 quasi_bpl) *m F0_isom
  by rewrite /= /D1_new_pert /D1_new /P1_new -mulmx_addl -mulmx_addr 
            -!mulmxA F0G0 mulmx1 mulmxA F1G1 mul1mx.
by rewrite -!mulmxA [F0_isom *m _]mulmxA G0F0 mul1mx.
Qed.

Lemma D1g1_g0d1_BPL : G1_BPL *m D1_BPL = D1'_BPL *m G0_BPL.
Proof. 
rewrite /G1_BPL /D1'_BPL /D1_BPL /G0_BPL mulmxA
        (reduction_definition.G0_morph quasi_bpl).
have -> : (D1 + P1) = G1_isom *m (reduction_definition.D1 quasi_bpl) *m F0_isom
  by rewrite /= /D1_new_pert /D1_new /P1_new -mulmx_addl -mulmx_addr 
            -!mulmxA F0G0 mulmx1 mulmxA F1G1 mul1mx.
by rewrite -!mulmxA [F1_isom *m _]mulmxA G1F1 mul1mx.
Qed.

Definition rho_bpl := 
 (Build_reduction (Logic.eq_sym d1f1_f0D1_BPL)  
   (Logic.eq_sym D1g1_g0d1_BPL) f0g0_BPL f1g1_BPL 
   g0f0_D1h0_BPL g1f1_D2h1_h0D1_BPL f1h0_BPL h0g0_BPL).

End reduction_properties.

End definitions_BPL.

End new_perturbation.

End perturbation.

End decomposition.

End BPL.



