Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun tuple ssralg matrix mxalgebra bigop.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


Section decomposition.

Open Scope ring_scope.

Variable K : fieldType.
Variables (c1 c2 d1 d2 : nat).
Variables (F1 : 'M[K]_(c1,d1)) (G1 : 'M[K]_(d1,c1)) (F2 : 'M[K]_(c2,d2))
          (G2 : 'M[K]_(d2,c2)) (D2 : 'M[K]_(c2,c1)) (D2': 'M[K]_(d2,d1))
          (H1 : 'M[K]_(c1,c2)). 

Hypothesis F_morph : (D2 *m F1) = (F2 *m D2').
Hypothesis G_morph : (D2' *m G1) = (G2 *m D2).
Hypothesis ax0 : (G1 *m F1) = 1%:M.
Hypothesis ax0' : (G2 *m F2) = 1%:M.
Hypothesis ax1  : (F1 *m G1) + (H1 *m D2) = 1%:M.
Hypothesis ax1' : (F2 *m G2) + (D2 *m H1) = 1%:M.
Hypothesis ax2 : (H1 *m F2) = 0.
Hypothesis ax3_0 : (H1 *m F2) = 0. 
Hypothesis ax4 : (G1 *m H1) = 0.


Section decomposition_C2.

Lemma h1d2_d1h0_in_kerf_C2 : (D2 *m H1) *m F2 = 0.
Proof.
by rewrite -!mulmxA ax2 mulmx0. 
Qed.

Lemma ker_intersection (m1 n1 n2: nat) (A:'M[K]_(m1,n1)) (B:'M[K]_(m1,n2)):
(kermx A :&: kermx B :=: kermx (row_mx A B))%MS.
Proof.
apply/eqmxP; apply/rV_eqP => u.
rewrite sub_capmx !(sameP sub_kermxP eqP) mul_mx_row -row_mx0.
by  apply/andP/eqP=> [[/eqP-> /eqP->] | /eq_row_mx[-> ->]]. 
Qed.


Lemma decomposition_C2' : (1%:M :=: (col_mx G2 (kermx F2))%MS)%MS.
Proof.
apply/eqmxP; apply/andP; split.
  apply/submxP.
  exists (row_mx F2 ((D2 *m H1) *m (col_ebase F2))).
  by rewrite mul_row_col (mulmxKV_ker h1d2_d1h0_in_kerf_C2). 
apply/submxP.
exists (col_mx G2 (kermx F2)).
by rewrite mulmx1.
Qed.

Lemma decomposition_C2 : (1%:M :=: ((kermx F2) + G2)%MS)%MS.
Proof.
by rewrite addsmxC;
apply : (eqmx_trans decomposition_C2' (eqmx_sym (addsmxE G2 (kermx F2)))).
Qed.

End decomposition_C2.


Section decomposition_C1.

Lemma h1d2_d1h0_in_kerf_C1 : (H1 *m D2) *m F1 = 0.
Proof.
by rewrite -!mulmxA F_morph mulmxA ax2 mul0mx. 
Qed.

Lemma decomposition_C1' : (1%:M :=: (col_mx G1 (kermx F1))%MS)%MS.
Proof.
apply/eqmxP; apply/andP; split.
  apply/submxP.
  exists (row_mx F1 ((H1 *m D2) *m (col_ebase F1))).
  by rewrite mul_row_col (mulmxKV_ker h1d2_d1h0_in_kerf_C1). 
apply/submxP.
exists (col_mx G1 (kermx F1)).
by rewrite mulmx1.
Qed.

Lemma decomposition_C1 : (1%:M :=: ((kermx F1) + G1)%MS)%MS.
Proof.
by rewrite addsmxC;
apply : (eqmx_trans decomposition_C1' (eqmx_sym (addsmxE G1 (kermx F1)))).
Qed.

End decomposition_C1.

Section auxiliary_results.

Lemma cast_rank1 (m n : nat) (M : 'M[K]_(m,n)) : 
 ((\rank M) + (m - (\rank M))%N )%N = m. 
Proof.
by rewrite (subnKC (rank_leq_row M)).
Qed.

Lemma ltnSadd (m n : nat) : m + n < m = false.
Proof.
elim : n => [|n IH]; first by rewrite addn0 ltnn.
rewrite addnS.
case H : ((m + n).+1 < m); last by done.
have H2 : m + n < (m + n).+1 by done.
move : (ltn_trans H2 H) => H3.
by rewrite H3 in IH.  
Qed.


Definition ker_min  (m n : nat) (M : 'M_(m,n)) :=
  (castmx ((Logic.eq_refl (m-\rank M)%N), (cast_rank1 M))
          (row_mx (@const_mx _ (m-\rank M) (\rank M) 0) 
                  1%:M)) *m 
  (kermx M).

Lemma castmx_copid (m m' r: nat) (eq_mm : (m = m') * (m = m'))
  (R : ringType) :
  castmx eq_mm (copid_mx r) = (@copid_mx R m' r).
Proof. 
by case: eq_mm; case: m' / => b; rewrite castmx_id.
Qed.

Lemma castmx_blockmx_copid (m m' r: nat) (eq_mm : ((r + m')%N = m) * ((r+m')%N = m))
  (R : ringType) : (castmx eq_mm (block_mx 0 0 0 1%:M)) = (@copid_mx R m r).
Proof.
case : eq_mm; case : m / => b; rewrite castmx_id.
apply/matrixP => i j; rewrite mxE; case : splitP => i0 H;
rewrite mxE; case : splitP => j0 Hj; 
rewrite !mxE {2}H ?(ltn_ord i0) ?Bool.andb_true_r ?ltnSadd ?Bool.andb_false_r
   ?GRing.subr0 /=. 
      case Hij : (@eq_op nat_eqType (@nat_of_ord (addn r m') i)
                    (@nat_of_ord (addn r m') j)).
         move : Hij; move/eqP => Hij;  move : (ord_inj Hij); 
         move/eqP => -> /=.
         have H11 : 1%:R == 1%:R by done.
         move : (H11 R) => H12.      
         rewrite -GRing.subr_eq0 in H12.
         by rewrite (eqP H12).
       case Hij2 : (i == j); last by rewrite GRing.subr0.
       move : (eqP Hij2) => Hij3.
       by rewrite Hij3 in Hij; move : Hij; move/eqP.
    case Hij: (@eq_op nat_eqType (@nat_of_ord (addn r m') i)(@nat_of_ord (addn r m') j)).
      move : Hij; move/eqP => Hij; move : (ord_inj Hij);  move/eqP => -> /=.
      have H11 : 1%:R == 1%:R by done.
      move : (H11 R) => H12.      
      rewrite -GRing.subr_eq0 in H12.
      by rewrite (eqP H12).
    case Hij2 : (i == j); last by rewrite GRing.subr0.
    move : (eqP Hij2) => Hij3.
    by rewrite Hij3 in Hij; move : Hij; move/eqP.
  case Hij: (i == j); last by done.
  move : (eqP Hij) => Hij2.
  rewrite Hij2 in H.
  by move : (ltn_ord j0); rewrite -Hj H ltnSadd.
case Hij: (i0 == j0). 
  move : (eqP Hij) => Hij2.
  rewrite Hij2 in H.
  have Hij3 : (@eq_op nat_eqType (@nat_of_ord (addn r m') i)
                    (@nat_of_ord (addn r m') j)) by rewrite H Hj.
  move: (ord_inj (eqP Hij3)) => ->.
  by have -> : (j == j); by done. 
case Hij2 : (i == j); last by done.
move : (eqP Hij2) => Hij3.
rewrite Hij3 Hj in H.
have Hij4 : (r + j0 - r)%N = (r + i0 - r)%N by rewrite H. 
rewrite -{2 4}(addn0 r) !subn_add2l !subn0 in Hij4.
move : (ord_inj Hij4) => Hij5.
rewrite Hij5 in Hij.
by move : Hij; move/eqP.
Qed.

Lemma mulmx_cast : forall (R : ringType) (m1 n1 m2 n2 m3 m4 : nat) (eq_mn1 : (m1 = m2) * (n1 = n2))
     (eq_mn2 : (n1 = n2) * (m3 = m4)) (eq_mn3 : (m1 = m2) * (m3 = m4)) (A : 'M[R]_(m1, n1)) (B :'M[R]_(n1,m3)), 
     ((castmx eq_mn1 A) *m (castmx eq_mn2 B)) = (castmx eq_mn3 (A *m B)).
Proof.
move => R m1 n1 m2 n2 m3 m4 [eq_mn1 eq_mn1'] [eq_mn2 eq_mn2'] [eq_mn3 eq_mn3'] A B.
move: (eq_mn1) (eq_mn1') (eq_mn2) (eq_mn2') (eq_mn3) (eq_mn3') A B. 
rewrite eq_mn1 eq_mn1' eq_mn3' => eq_mn_0 eq_mn_1 eq_mn_2 eq_mn_3 eq_mn_4 eq_mn_5 A B.
by rewrite !castmx_id.
Qed.


Lemma ker_min_kermx (m n : nat) (M : 'M_(m,n)) : 
  (kermx M :=: (ker_min M))%MS. 
Proof.
apply/eqmxP/andP; split; apply/submxP; last first.
  by exists (castmx ((Logic.eq_refl (m-\rank M)%N), (cast_rank1 M))
                    (row_mx (@const_mx _ (m-\rank M) (\rank M) 0) 1%:M)).
exists (castmx ((cast_rank1 M),(Logic.eq_refl (m-\rank M)%N))
               (col_mx (@const_mx _ (\rank M) (m-\rank M)%N 0) 1%:M)).
rewrite /ker_min mulmxA mulmx_cast ?cast_rank1 //.
move => Hyp0.
by rewrite mul_col_row !mul0mx mulmx0 mulmx1 /kermx castmx_blockmx_copid 
           mulmxA copid_mx_id ?(rank_leq_row M). 
Qed.

End auxiliary_results.



Section isomorphism.

Definition castF1 := 
(castmx ((cast_rank1 F1), (Logic.eq_refl (c1-\rank F1)%N))
    (col_mx (@const_mx K (\rank F1) (c1-\rank F1)%N 0) 1%:M)).

Definition castF2 := 
(castmx ((cast_rank1 F2), (Logic.eq_refl (c2-\rank F2)%N))
    (col_mx (@const_mx K (\rank F2) (c2-\rank F2)%N 0) 1%:M)).

Definition G1_isom := (col_mx (ker_min F1) G1).
Definition F1_isom := (row_mx ((H1 *m D2) *m (col_ebase F1) *m castF1) F1).
Definition G2_isom := (col_mx (ker_min F2) G2).
Definition F2_isom := (row_mx ((D2 *m H1) *m (col_ebase F2) *m castF2) F2).
Definition D2_change_base:= G2_isom *m D2 *m F1_isom.
Definition H1_change_base:= G1_isom *m H1 *m F2_isom.

Lemma F1G1_isom : F1_isom *m G1_isom = 1%:M.
Proof.
rewrite mul_row_col GRing.addrC -!mulmxA /ker_min
       [castF1 *m (_ *m _)]mulmxA mulmx_cast ?cast_rank1 //. 
move => Hyp0.
rewrite mul_col_row !mul0mx mulmx0 mul1mx castmx_blockmx_copid /kermx
       [copid_mx (_) *m (_ *m _)]mulmxA copid_mx_id 
       ?(rank_leq_row F1) //.
have -> : (copid_mx (\rank F1) *m invmx (col_ebase F1)) = kermx F1 by done.
by rewrite mulmxA mulmxA (mulmxKV_ker h1d2_d1h0_in_kerf_C1).
Qed.

Lemma G1F1_isom : 
  G1_isom *m F1_isom = 1%:M.
Proof.
rewrite mul_col_row ax0 ![G1 *m (_ *m _)]mulmxA ax4 !mul0mx.
have H : (ker_min F1 *m F1) = 0.
  move : (ker_min_kermx F1).
  move/eqmxP; move/andP => [_ Hb].
  move : Hb; move/submxP => [D'' HD''].
  by rewrite HD'' -mulmxA mulmx_ker mulmx0.
have -> : (H1 *m D2) = 1%:M - (F1 *m G1).
  rewrite -ax1 [(_ *m _) + (_ *m _)]GRing.addrC -GRing.addrA.
  have H3 : F1 *m G1 == F1 *m G1 by done.
  rewrite -GRing.subr_eq0 in H3.
  by rewrite (eqP H3) GRing.addr0.
rewrite !mulmx_addl !mulNmx mul1mx mulmx_addr mulmxN !mulmxA H !mul0mx
        GRing.subr0 /ker_min /kermx -!mulmxA mulmxA.
have -> : (invmx (col_ebase F1) *m (col_ebase F1 *m castF1)) = castF1 
  by rewrite mulmxA (mulVmx (col_ebase_unit F1)) mul1mx.
rewrite -(@castmx_blockmx_copid _ (c1 - \rank F1)) ?(cast_rank1 F1) // => Hyp0; rewrite mulmx_cast ?(cast_rank1 F1) // => Hyp1; rewrite mulmx_cast ?(cast_rank1 F1) // => Hyp2.
by rewrite mul_row_block mul_row_col mul0mx GRing.add0r mulmx0 GRing.add0r
          !mulmx1 mulmx0 GRing.add0r castmx_id -scalar_mx_block.
Qed.

Lemma F2G2_isom : F2_isom *m G2_isom = 1%:M.
Proof.
rewrite mul_row_col GRing.addrC -!mulmxA /ker_min
       [castF2 *m (_ *m _)]mulmxA mulmx_cast ?cast_rank1 //. 
move => Hyp0.
rewrite mul_col_row !mul0mx mulmx0 mul1mx castmx_blockmx_copid /kermx
       [copid_mx (_) *m (_ *m _)]mulmxA copid_mx_id 
       ?(rank_leq_row F2) //.
have -> : (copid_mx (\rank F2) *m invmx (col_ebase F2)) = kermx F2 by done.
by rewrite mulmxA mulmxA (mulmxKV_ker h1d2_d1h0_in_kerf_C2).
Qed.

Lemma G2F2_isom : G2_isom *m F2_isom = 1%:M.
Proof.
rewrite mul_col_row ax0' ![G2 *m (_ *m _)]mulmxA -G_morph 
        -[_ *m G1 *m _]mulmxA ax4 !mulmx0 !mul0mx.
have H : (ker_min F2 *m F2) = 0.
  move : (ker_min_kermx F2).
  move/eqmxP; move/andP => [_ Hb].
  move : Hb; move/submxP => [D'' HD''].
  by rewrite HD'' -mulmxA mulmx_ker mulmx0.
have -> : (D2 *m H1) = 1%:M - (F2 *m G2).
  rewrite -ax1' [(_ *m _) + (_ *m _)]GRing.addrC -GRing.addrA.
  have H3 : F2 *m G2 == F2 *m G2 by done.
  rewrite -GRing.subr_eq0 in H3.
  by rewrite (eqP H3) GRing.addr0.
rewrite !mulmx_addl !mulNmx mul1mx mulmx_addr mulmxN !mulmxA H !mul0mx
        GRing.subr0 /ker_min /kermx -!mulmxA mulmxA.
have -> : (invmx (col_ebase F2) *m (col_ebase F2 *m castF2)) = castF2 
  by rewrite mulmxA (mulVmx (col_ebase_unit F2)) mul1mx.
rewrite -(@castmx_blockmx_copid _ (c2 - \rank F2)) ?(cast_rank1 F2) // => Hyp0; rewrite mulmx_cast ?(cast_rank1 F2) // => Hyp1; rewrite mulmx_cast ?(cast_rank1 F2) // => Hyp2.
by rewrite mul_row_block mul_row_col mul0mx GRing.add0r mulmx0 GRing.add0r
          !mulmx1 mulmx0 GRing.add0r castmx_id -scalar_mx_block.
Qed.



Lemma D2_change_baseP : D2_change_base = 
 block_mx (ker_min F2 *m D2 *m (H1 *m D2 *m col_ebase F1 *m castF1))
           0 0 D2'.
Proof.
rewrite /D2_change_base mul_col_mx mul_col_row.
have -> : (G2 *m D2 *m F1) = D2' by rewrite -G_morph -mulmxA ax0 mulmx1.
have -> : (ker_min F2 *m D2 *m F1) = 0.
  move : (ker_min_kermx F2).
  move/eqmxP; move/andP => [_ Hb].
  move : Hb; move/submxP => [D'' HD''].
  by rewrite HD'' -mulmxA F_morph -mulmxA [_ *m (F2 *m _)]mulmxA mulmx_ker 
             mul0mx mulmx0. 
have -> : (G2 *m D2 *m (H1 *m D2 *m col_ebase F1 *m castF1)) = 0.
  by rewrite !mulmxA -G_morph -!mulmxA [G1 *m (_ *m _)]mulmxA ax4 mul0mx 
             mulmx0.
by done.
Qed.

Lemma H1_change_baseP : H1_change_base = 
 block_mx (ker_min F1 *m H1 *m (D2 *m H1 *m col_ebase F2 *m castF2)) 0 0 0.
Proof.
rewrite /H1_change_base mul_col_mx mul_col_row.
have -> : (G1 *m H1 *m F2) = 0 by rewrite ax4 mul0mx.
have -> : (ker_min F1 *m H1 *m F2) = 0 by rewrite -mulmxA ax2 mulmx0.
have -> : (G1 *m H1 *m (D2 *m H1 *m col_ebase F2 *m castF2)) = 0.
  by rewrite ax4 !mul0mx.
by rewrite !mulmxA.
Qed.

Lemma D2_H2_id: 
   ulsubmx (D2_change_base *m H1_change_base) = 1%:M.
Proof.
rewrite /D2_change_base /H1_change_base -mulmxA 
       ![F1_isom *m (_ *m _)]mulmxA F1G1_isom mul1mx -mulmxA
       [D2 *m (_ *m _)]mulmxA.
have -> : (D2 *m H1) = 1%:M - (F2 *m G2).
  rewrite -ax1' [(_ *m _) + (_ *m _)]GRing.addrC -GRing.addrA.
  have H3 : F2 *m G2 == F2 *m G2 by done.
  rewrite -GRing.subr_eq0 in H3.
  by rewrite (eqP H3) GRing.addr0.
rewrite mulmx_addl mulmx_addr mulNmx mulmxN mul1mx G2F2_isom mulmxA 
        /G2_isom /F2_isom mul_col_mx mul_mx_row mulmxA !mul_col_mx
        -block_mxEh scalar_mx_block opp_block_mx add_block_mx block_mxKul
        !mulmxA.
have -> : (ker_min F2 *m F2) = 0.
  move : (ker_min_kermx F2).
  move/eqmxP; move/andP => [_ Hb].
  move : Hb; move/submxP => [D'' HD''].
  by rewrite HD'' -mulmxA mulmx_ker mulmx0.
by rewrite !mul0mx GRing.subr0.
Qed.

Lemma D2_H2_block: 
(ker_min F2 *m D2 *m (H1 *m D2 *m col_ebase F1 *m castF1)) *m 
(ker_min F1 *m H1 *m (D2 *m H1 *m col_ebase F2 *m castF2)) = 1%:M.
Proof.
move : D2_H2_id.
by rewrite D2_change_baseP H1_change_baseP mulmx_block block_mxKul mul0mx
           GRing.addr0. 
Qed.

End isomorphism.

End decomposition.
