Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun tuple ssralg matrix mxalgebra bigop.
Require Import reduction_definition.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


(* Hexagonal lemma for two matrices *)

Section hexagonal_lemma.

Local Open Scope ring_scope.

Variable K : fieldType.

Variables (m1 m2 n1 n2: nat).
Variables (D1 : 'M[K]_(m1+m2,n1+n2)).

Section descomposition_D1.

Definition daa := (ulsubmx D1).
Definition dba := (ursubmx D1).
Definition dab := (dlsubmx D1).
Definition dbb := (drsubmx D1).


Lemma decomposition_D1 : D1 =
  block_mx daa dba dab dbb.
Proof.
by rewrite -(submxK D1). 
Qed.

End descomposition_D1.


Section definitions.

Variable daa_1 : 'M[K]_(n1,m1).

Hypothesis daa_invertible : 
  daa_1 *m daa = 1%:M /\ daa *m daa_1 = 1%:M.

Definition d1 := dbb - (dab *m daa_1 *m dba).
Definition f0 := (col_mx (- daa_1 *m dba) 1%:M).
Definition g0 := (row_mx 0 1%:M):'M[K]_(n2,n1+n2).
Definition f1 := (col_mx 0 1%:M) : 'M[K]_(m1+m2,m2).
Definition g1 := (row_mx (- dab *m daa_1) 1%:M).
Definition h0 := (block_mx daa_1 0 0 0) : 'M_(n1+n2,m1+m2).

Section reduction_properties.

Lemma f0g0 : g0 *m f0 = 1%:M.
Proof.
by rewrite mul_row_col mul0mx mul1mx GRing.add0r.
Qed.

Lemma f1g1 : g1 *m f1 = 1%:M.
Proof.
by rewrite !mul_row_col !mulmx0 mul1mx !GRing.add0r.
Qed.

Lemma f1h0 : h0 *m f1 = 0.
Proof.
by rewrite mul_block_col !mulmx0 !mul0mx !GRing.addr0 col_mx0.
Qed.

Lemma h0g0 : g0 *m h0 = 0.
Proof.
by rewrite mul_row_block !mulmx0 !mul0mx !GRing.addr0 row_mx0. 
Qed.

Lemma g0f0_D1h0 : f0 *m g0 + h0 *m D1 = 1%:M.
Proof.
rewrite mul_col_row decomposition_D1 mulmx_block add_block_mx !mulmx0
        !mul0mx !GRing.addr0 !mulmx1 !GRing.add0r.
move : daa_invertible => [-> _].
have -> : (- daa_1 *m dba + daa_1 *m dba) = 0.
  rewrite GRing.addrC mulNmx.
  have H : daa_1 *m dba == daa_1 *m dba by done.
  rewrite -GRing.subr_eq0 in H.
  by rewrite (eqP H).
by rewrite scalar_mx_block.
Qed.

Lemma m_m_0 (R : ringType) (t1 t2 : nat) (M : 'M[R]_(t1,t2)) : M - M = 0.
Proof.
by move : (eq_refl M); rewrite -GRing.subr_eq0; move/eqP => ->.
Qed.

Lemma g1f1_D2h1_h0D1 : f1 *m g1 + D1 *m h0 = 1%:M.
Proof. 
rewrite decomposition_D1 !mul_col_row !mulmx_block
        !add_block_mx !mulmx1 !mul0mx !mulmx0 !GRing.addr0 mul1mx
        GRing.add0r.
move : daa_invertible => [_ ->].
have -> : (- dab *m daa_1 + dab *m daa_1) = 0.
  by rewrite GRing.addrC mulNmx m_m_0.
by rewrite -scalar_mx_block.
Qed.

Lemma d1f1_f0D1 : f1 *m d1 = D1 *m f0.
Proof. 
rewrite decomposition_D1 mul_col_mx mul_block_col !mulmx1
        mul0mx.
have -> : (daa *m (- daa_1 *m dba) + dba) = 0.
  rewrite mulmxA mulmxN.
  move : daa_invertible => [_ ->].
  by rewrite mulNmx mul1mx GRing.addrC m_m_0.
rewrite mul1mx.
by have -> : d1 = (dab *m (- daa_1 *m dba) + dbb); first
  by rewrite /d1 mulNmx mulmxN GRing.addrC mulmxA.
Qed.

Lemma D1g1_g0d1 : g1 *m D1 = d1 *m g0.
Proof. 
rewrite decomposition_D1 mul_row_block !mul_mx_row !mulmx0 !mulmx1 
        !mul1mx !mulNmx -mulmxA.
move : daa_invertible => [-> _].
by rewrite mulmx1 GRing.addrC m_m_0 GRing.addrC.
Qed.


End reduction_properties.

Definition rho := (Build_reduction  (Logic.eq_sym d1f1_f0D1)
   (Logic.eq_sym D1g1_g0d1) f0g0 f1g1 g0f0_D1h0 g1f1_D2h1_h0D1
   f1h0 h0g0). 

End definitions.

End hexagonal_lemma.

