Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun tuple ssralg matrix mxalgebra bigop.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


Section invmx.

Import GRing.Theory.
Local Open Scope ring_scope.

Variable R : comUnitRingType.

Lemma invmx_left : forall m (M M' : 'M[R]_m), M *m M' = 1%:M -> M' *m M = 1%:M.
Proof.
move=> m M M' H.
have hM : (M \in unitmx) by case: (mulmx1_unit H).
have h' : (M' = invmx M) by rewrite -(mulKmx hM M') H mulmx1.
by rewrite h' mulVmx.
Qed.

Lemma invmx_uniq m (M M' : 'M[R]_m) :
  M *m M' = 1%:M -> M' = invmx M.
Proof.
move=> H.
have hM : (M \in unitmx) by case: (mulmx1_unit H).
by rewrite -[M']mulmx1 -(mulmxV hM) mulmxA (invmx_left H) mul1mx.
Qed.

End invmx.

Section pot_matrix.

Import GRing.Theory.
Local Open Scope ring_scope.

Variable (n : nat) (R : comUnitRingType).

Fixpoint pot_matrix (M : 'M[R]_n) (m : nat) : 'M[R]_n :=
match m with
| 0 => 1%:M
| S p => M *m (pot_matrix M p)
end.

Lemma pot_1 (i : nat) (M : 'M[R]_n) : 
   (pot_matrix M i *m M) =  (pot_matrix M i.+1).
Proof.
elim : i; first by rewrite mul1mx /pot_matrix mulmx1. 
move => n0 IH; rewrite {1}/pot_matrix -/pot_matrix.
by rewrite -mulmxA IH.
Qed.

End pot_matrix.



Section canonical_matrix.

Open Scope ring_scope.

Variables (R : comUnitRingType) (n : nat).

Canonical matrix_monoid :=
  Monoid.Law (@mulmxA R n n n n) (@mul1mx R n n) (@mulmx1 R n n).
Canonical matrix_muloid :=
  Monoid.MulLaw (@mul0mx R n n n) (@mulmx0 R n n n).
Canonical matrix_addoid := 
  Monoid.AddLaw (@mulmx_addl R n n n) (@mulmx_addr R n n n).

End canonical_matrix.


Section inverse_nilpotent.

Open Scope ring_scope.

Variables (R : comUnitRingType) (n : nat).

Lemma submx_sub (n0 : nat) (m1 n1 p1 : 'M[R]_n0) :
 m1 - n1 - p1 = m1 - (n1 + p1).
Proof.
apply/eqP.
by rewrite !GRing.subr_eq -GRing.addrA [p1+n1]GRing.addrC GRing.subrK.
Qed.

Lemma inverse_I_minus_M : forall (M : 'M[R]_n), 
 (exists m, pot_matrix M m = 0) -> exists N, N *m (1%:M - M) = 1%:M.
Proof.
case : n; first by exists 0; rewrite !thinmx0.
move => n0 M [m].
case : m.
 rewrite /pot_matrix => H.
 have : (@determinant R n0.+1 1%:M) = (@determinant R n0.+1 0) by rewrite H. rewrite det1 det0 => H1; move : (@GRing.nonzero1r R).
 by rewrite H1; move/eqP.
move => n1.
exists (\sum_(0<=i<n1.+1) (pot_matrix M i)).
rewrite (@big_distrl _ 0 _ (@matrix_addoid R n0.+1)).
have -> : \big[matrix_addoid R n0.+1/0]_(0 <= i < n1.+1)
        (matrix_muloid R n0.+1) (pot_matrix M i) (1%:M - M) =
      \big[matrix_addoid R n0.+1/0]_(0 <= i < n1.+1)
        ((pot_matrix M i) *m (1%:M - M)) by done.
have H : forall i : nat , (fun _ => true) i ->
   (fun i => (pot_matrix M i *m (1%:M - M))) i =
   (fun i => ((pot_matrix M i) - ((pot_matrix M i) *m M))) i
  by move => i;
    rewrite [(pot_matrix M i) *m (1%:M - M)]mulmx_subr mulmx1.
rewrite (@eq_bigr _ _ _ _ _ _ _ _ H).  
clear H.
rewrite GRing.Theory.sumr_sub.
have H : forall i : nat , (fun _ => true) i ->
   (fun i => (pot_matrix M i *m M)) i =
   (fun i => (pot_matrix M i.+1)) i.
 by move => i _; rewrite pot_1.
rewrite (@eq_bigr _ _ _ _ _ _ _ _ H) big_nat_recl big_nat_recr
        -GRing.addrA.  
have -> :
  (GRing.add_monoid (matrix_zmodType R n0.+1 n0.+1))
     (\big[GRing.add_monoid (matrix_zmodType R n0.+1 n0.+1)/0]_(0 <= i < n1)
          pot_matrix M i.+1) (pot_matrix M n1.+1) =
  (\big[GRing.add_monoid (matrix_zmodType R n0.+1 n0.+1)/0]_(0 <= i < n1)
          pot_matrix M i.+1) + (pot_matrix M n1.+1) by done.
rewrite -submx_sub.
have H1 : \big[+%R/0]_(0 <= i < n1) pot_matrix M i.+1 == 
   \big[GRing.add_monoid (matrix_zmodType R n0.+1 n0.+1)/0]_(0 <= i < n1)
       pot_matrix M i.+1 by done.
rewrite -GRing.subr_eq0 in H1.
by rewrite (eqP H1) GRing.sub0r {1}/pot_matrix p GRing.subr0.
Qed.

Lemma inverse_I_minus_M' : forall (M : 'M[R]_n), 
 (exists m, pot_matrix M m = 0) -> exists N, (1%:M - M) *m N = 1%:M.
Proof.
move => M H; move : (inverse_I_minus_M H) => [N eH].
exists N.
by rewrite (invmx_left eH).
Qed. 

Variable m : nat.
Variable (M : 'M[R]_n).
Hypothesis nilpotency_m : (pot_matrix M m = 0).

 
Lemma inverse_I_minus_M_big : 
   (\sum_(0<=i<m) (pot_matrix M i)) *m (1%:M - M) = 1%:M.
Proof.
move : M nilpotency_m.
case : n; first by move => M0 nilpotency_m0; rewrite !thinmx0. 
move => n0 M0.
case : m.
 rewrite /pot_matrix => H.
 have : (@determinant R n0.+1 1%:M) = (@determinant R n0.+1 0) by rewrite H. rewrite det1 det0 => H1; move : (@GRing.nonzero1r R).
 by rewrite H1; move/eqP.
move => n1 H1.
rewrite (@big_distrl _ 0 _ (@matrix_addoid R n0.+1)).
have -> : \big[matrix_addoid R n0.+1/0]_(0 <= i < n1.+1)
        (matrix_muloid R n0.+1) (pot_matrix M0 i) (1%:M - M0) =
      \big[matrix_addoid R n0.+1/0]_(0 <= i < n1.+1)
        ((pot_matrix M0 i) *m (1%:M - M0)) by done.
have H : forall i : nat , (fun _ => true) i ->
   (fun i => (pot_matrix M0 i *m (1%:M - M0))) i =
   (fun i => ((pot_matrix M0 i) - ((pot_matrix M0 i) *m M0))) i
  by move => i;
    rewrite [(pot_matrix M0 i) *m (1%:M - M0)]mulmx_subr mulmx1.
rewrite (@eq_bigr _ _ _ _ _ _ _ _ H).  
clear H.
rewrite GRing.Theory.sumr_sub.
have H : forall i : nat , (fun _ => true) i ->
   (fun i => (pot_matrix M0 i *m M0)) i =
   (fun i => (pot_matrix M0 i.+1)) i.
 by move => i _; rewrite pot_1.
rewrite (@eq_bigr _ _ _ _ _ _ _ _ H) big_nat_recl big_nat_recr
        -GRing.addrA.  
have -> :
  (GRing.add_monoid (matrix_zmodType R n0.+1 n0.+1))
     (\big[GRing.add_monoid (matrix_zmodType R n0.+1 n0.+1)/0]_(0 <= i < n1)
          pot_matrix M0 i.+1) (pot_matrix M0 n1.+1) =
  (\big[GRing.add_monoid (matrix_zmodType R n0.+1 n0.+1)/0]_(0 <= i < n1)
          pot_matrix M0 i.+1) + (pot_matrix M0 n1.+1) by done.
rewrite -submx_sub.
have HH1 : \big[+%R/0]_(0 <= i < n1) pot_matrix M0 i.+1 == 
   \big[GRing.add_monoid (matrix_zmodType R n0.+1 n0.+1)/0]_(0 <= i < n1)
       pot_matrix M0 i.+1 by done.
rewrite -GRing.subr_eq0 in HH1.
by rewrite (eqP HH1) GRing.sub0r {1}/pot_matrix H1 GRing.subr0.
Qed.

Lemma inverse_I_minus_M'_big :  
 (1%:M - M) *m (\sum_(0<=i<m) (pot_matrix M i)) = 1%:M.
Proof.
move : (inverse_I_minus_M_big) => eH.
by rewrite (invmx_left eH).
Qed. 

End inverse_nilpotent.

Section inverse_plus_nilpotent.

Open Scope ring_scope.

Lemma subr_opp (R : ringType) (t1 t2 : nat) (m n : 'M[R]_(t1,t2)) : 
  m - - n = m + n.
Proof.
apply/eqP.
rewrite !GRing.subr_eq -GRing.addrA. 
move : (eq_refl n) => H.
rewrite -GRing.subr_eq0 in H.
by rewrite (eqP H) GRing.addr0.
Qed.

Variables (R : comUnitRingType) (n : nat).

Variable m : nat.
Variable (M : 'M[R]_n).


Lemma pot_minus : (pot_matrix (- M) m = pot_matrix M m)
\/ (pot_matrix (- M) m = - pot_matrix M m).
Proof.
move : M.
case : n; first by move => M0; rewrite !thinmx0; left.
move => n0 M0.
elim : m => [/=| m1 HH]; first by left.
move : HH => [HH1 | HH2].
  right.
  have -> :  pot_matrix (- M0) m1.+1 = (- M0) *m pot_matrix (- M0) m1 by done.
  by rewrite HH1 mulNmx.
left.
have -> :  pot_matrix (- M0) m1.+1 = (- M0) *m pot_matrix (- M0) m1 by done.
by rewrite HH2 mulNmx mulmxN -GRing.sub0r subr_opp GRing.add0r.
Qed.

Lemma nilpotency_minus_m : (pot_matrix M m = 0) -> pot_matrix (- M) m = 0.
Proof.
move => nilpotency_m.
by move : pot_minus => [-> | ->]; rewrite nilpotency_m ?GRing.oppr0.
Qed.


Lemma inverse_I_plus_M_big : (pot_matrix M m = 0) ->
   (\sum_(0<=i<m) (pot_matrix (- M) i)) *m (1%:M + M) = 1%:M.
Proof.
move => nilpotency_m.
move : (inverse_I_minus_M_big (nilpotency_minus_m nilpotency_m)).
by rewrite subr_opp.
Qed.

Lemma inverse_I_plus_M_big' : (pot_matrix M m = 0) ->
   (1%:M + M) *m (\sum_(0<=i<m) (pot_matrix (- M) i)) = 1%:M.
Proof.
move => nilpotency_m.
move : (inverse_I_plus_M_big nilpotency_m) => eH.
by rewrite (invmx_left eH).
Qed. 

End inverse_plus_nilpotent.

